import unittest
import time
import sys
import subprocess
import requests
import json
from fam.utils import couchbase_utils

from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monarch, NAMESPACE
from fam.mapper import ClassMapper

from fam.database import CouchDBWrapper, SyncGatewayWrapper

COUCHDB_URL = "http://localhost:5984"
COUCHDB_NAME = "test"

REMOTE_URL ="http://paul:password@localhost:4984/sync_gateway"

SYNC_GATEWAY_HOST = "localhost"
SYNC_GATEWAY_PORT = "4984"
SYNC_GATEWAY_ADMIN_PORT = "4985"
SYNC_GATEWAY_NAME = "sync_gateway"

if sys.platform.startswith('linux'):
    SYNC_GATEWAY_PATH = "/opt/couchbase-sync-gateway/bin/sync_gateway"
else:
    SYNC_GATEWAY_PATH = "sync_gateway"


class SingleReplicationTests(unittest.TestCase):

        def setUp(self):

            mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monarch])

            #a new walrus sync gateway
            cmd = "{} -log HTTP+,Changes+,Events+    -url walrus:".format(SYNC_GATEWAY_PATH)

            time.sleep(0.5)
            self.gateway = subprocess.Popen(cmd, shell=True)
            time.sleep(0.25)

            url = "http://%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_ADMIN_PORT)
            self.sync_gateway = SyncGatewayWrapper(mapper, url, SYNC_GATEWAY_NAME)

            couchbase_utils.add_person_to_gateway(url, SYNC_GATEWAY_NAME, "test_person", "paul", "password", admin_channels=["public"])

            # and set up a new couchdb replicating to it
            self.couchdb = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, remote_url=REMOTE_URL)


        def tearDown(self):

            self.couchdb.clear_all_replications()

            # stop the gateway
            print "tear_down"
            self.gateway.kill()


        def test_put_into_couchdb(self):
            dog = Dog(name="woofer", channels=["*"])
            dog.save(self.couchdb)
            self.assertEqual(dog.namespace, NAMESPACE)
            self.assertEqual(dog.type, "dog")
            self.assertEqual(dog.name, "woofer")
            self.assertEqual(dog.__class__, Dog)
            self.assertNotEqual(dog.key, None)


        def test_put_into_sync_gateway(self):
            dog = Dog(name="woofer")
            dog.save(self.sync_gateway)
            self.assertEqual(dog.namespace, NAMESPACE)
            self.assertEqual(dog.type, "dog")
            self.assertEqual(dog.name, "woofer")
            self.assertEqual(dog.__class__, Dog)
            self.assertNotEqual(dog.key, None)


        def test_single_replication_up(self):
            dog = Dog(key="test_dog", name="woofer")
            dog.save(self.couchdb)
            sync_dog = self.sync_gateway.get("test_dog")
            self.assertEqual(sync_dog, None)
            self.couchdb.sync_up()
            # time.sleep(2)
            sync_dog = self.sync_gateway.get("test_dog")
            self.assertNotEqual(sync_dog, None)


        def test_single_replication_down(self):
            dog = Dog(key="test_dog", name="woofer", channels=["public"])
            dog.save(self.sync_gateway)
            self.couchdb.sync_down()
            # time.sleep(2)
            sync_dog = self.couchdb.get("test_dog")
            self.assertNotEqual(sync_dog, None)


class ContinuousReplicationTests(unittest.TestCase):

        def setUp(self):

            mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monarch])

            #a new walrus sync gateway
            cmd = "{} -log HTTP+,Changes+,Events+    -url walrus:".format(SYNC_GATEWAY_PATH)

            time.sleep(0.25)
            self.gateway = subprocess.Popen(cmd, shell=True)
            time.sleep(0.25)

            url = "http://%s:%s" % (SYNC_GATEWAY_HOST, SYNC_GATEWAY_ADMIN_PORT)
            self.sync_gateway = SyncGatewayWrapper(mapper, url, SYNC_GATEWAY_NAME)

            couchbase_utils.add_person_to_gateway(url, SYNC_GATEWAY_NAME, "test_person", "paul", "password", admin_channels=["public"])

            # and set up a new couchdb replicating to it
            self.couchdb = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True, remote_url=REMOTE_URL, continuous=True)


        def tearDown(self):

            self.couchdb.clear_all_replications()

            # stop the gateway
            print "tear_down"
            self.gateway.kill()


        def test_single_replication_up(self):
            dog = Dog(key="test_dog", name="woofer")
            dog.save(self.couchdb)
            time.sleep(1)
            sync_dog = self.sync_gateway.get("test_dog")
            self.assertNotEqual(sync_dog, None)


        def test_single_replication_down(self):
            dog = Dog(key="test_dog", name="woofer", channels=["public"])
            dog.save(self.sync_gateway)
            time.sleep(1)
            sync_dog = self.couchdb.get("test_dog")
            self.assertNotEqual(sync_dog, None)





import unittest
from fam.tests.models.test01 import Monster, Weapons

class AnythingBaseTests:

    class AnythingTests(unittest.TestCase):

        db = None

        def test_Object_field(self):

            weapons = Weapons("large", "hot", ["2", "2"])
            key = "a_monster"
            monster = Monster(key=key, weapons=weapons, name="bill")
            self.db.put(monster)
            got_monster = self.db.get(key)
            self.assertTrue(isinstance(got_monster.weapons, Weapons))




import unittest
from decimal import Decimal
from fractions import Fraction
from fam.extra_types.lat_long import LatLong
from google.cloud.firestore_v1beta1 import GeoPoint

import os
import datetime
import pytz

from fam.database.firestore_adapter import FirestoreDataAdapter



import fam
from fam.exceptions import *
from fam.tests.models.test04 import Fish
from fam.tests.models.test04 import NAMESPACE as fish_namespace
from fam.tests.models.test01 import Dog
from fam.tests.models.test01 import  NAMESPACE as dog_namespace

from fam.database import FirestoreWrapper
from fam.mapper import ClassMapper

SECRETS_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "secrets")
DATA_DIR = os.path.join(os.path.dirname(fam.__file__), "tests", "data")


class TestSerialisation(unittest.TestCase):

    """
    str
    list
    dict
    bool
    float
    int
    Decimal
    Fraction
    LatLong
    datetime
    bytes
    unicode/str utf8
    """


    def setUp(self):
        self.adapter = FirestoreDataAdapter()


    def test_serialise_basic_types(self):

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)

        # these should all pass through unchanged
        self.assertEquals(doc, serialised)


    def test_serialise_numerics(self):

        doc = {"name": "Paul",
               "age": 53,
               "height": Decimal("5.9"),
               "fraction": Fraction(1, 2),
               "is_stoopid": True,
               "children": ["Sol", "Jake"],
               "favorites": {"drink": "coffee", "food": Decimal("4.2")}
               }

        serialised = self.adapter.serialise(doc)


        self.assertEquals(serialised["height"], "::decimal::5.9")
        self.assertEquals(serialised["fraction"], "::fraction::1/2")
        self.assertEquals(serialised["favorites"]["food"], "::decimal::4.2")


    def test_serialise_latlong_datetime(self):

        birthday = datetime.datetime(1964, 12, 5)

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "location": LatLong(51.5102213, -0.1178892),
                "birthday": birthday,
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)

        self.assertEquals(serialised["birthday"], birthday)
        self.assertTrue(isinstance(serialised["location"], GeoPoint))
        self.assertEquals(serialised["location"].longitude, -0.1178892)
        self.assertEquals(serialised["location"].latitude, 51.5102213)


    def test_serialise_bytes(self):

        doc = { "name": "Paul",
                "age": 53,
                "height": 5.9,
                "is_stoopid": True,
                "birthday": b"aq3restdyrgvdhrjb",
                "children": ["Sol", "Jake"],
                "favorites": {"drink": "coffee", "food": "egg and chips"}
        }

        serialised = self.adapter.serialise(doc)

        self.assertTrue(isinstance(serialised["birthday"], bytes))



class TestOptimiseSerialisationDatabase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        creds_path = os.path.join(SECRETS_DIR, "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([Dog])
        cls.db = FirestoreWrapper(mapper, creds_path, namespace=dog_namespace)
        cls.clear_db()


    @classmethod
    def clear_db(cls):
        cls.db.delete_all("dog")


    def test_data_base_id(self):

        dog = Dog.create(self.db, name="woofer")

        dog_id = dog.key

        self.assertTrue(dog.key is not None)

        doc_ref = self.db.db.collection("dog").document(dog_id)

        doc = doc_ref.get()
        as_dict = doc.to_dict()

        # self.assertTrue("_id" not in as_dict)
        self.assertTrue("type" not in as_dict)
        self.assertTrue("namespace" not in as_dict)

        got_dog = Dog.get(self.db, dog_id)

        self.assertTrue(got_dog.key == dog_id)
        self.assertTrue(got_dog.namespace == dog.namespace)
        self.assertTrue(got_dog.type == "dog")




class TestDatabase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        creds_path = os.path.join(SECRETS_DIR, "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([Fish])
        cls.db = FirestoreWrapper(mapper, creds_path, namespace=fish_namespace)
        cls.clear_db()


    @classmethod
    def clear_db(cls):
        cls.db.delete_all("fish")



    def test_fish(self):

        loc = LatLong(51.2345, -1.4533)
        birthday = datetime.datetime(1964, 12, 5, tzinfo=pytz.utc)

        image_path = os.path.join(DATA_DIR, "goldfish.jpg")
        with open(image_path, "rb") as f:
            image_data = f.read()

        fish = Fish.create(self.db, name="Nemo",
                           location=loc,
                           born=birthday,
                           length=Decimal("45.7"),
                           edible_fraction=Fraction(1, 3),
                           image=image_data
        )

        fish2 = Fish.get(self.db, fish.key)

        self.assertTrue(type(fish2.location) == LatLong)
        self.assertEqual(fish2.location.latitude, 51.2345)

        self.assertTrue(isinstance(fish2.born, datetime.datetime))
        self.assertEqual(fish2.born, birthday)

        self.assertTrue(isinstance(fish2.length, Decimal))
        self.assertEqual(fish2.length, Decimal("45.7"))

        self.assertTrue(isinstance(fish2.edible_fraction, Fraction))
        self.assertEqual(fish2.edible_fraction, Fraction(1, 3))

        self.assertTrue(isinstance(fish2.image, bytes))
        self.assertEqual(fish2.image, image_data)










from fam.blud import GenericObject
from fam.fields import *


NAMESPACE = "glowinthedark.co.uk/test05"


class Cat(GenericObject):
    fields = {
        "name": StringField(),
        "colour": StringField(immutable=True),
        "tail": BoolField(immutable=True, default=True),
        "legs": NumberField(required=True),
        "owner_id": ReferenceTo(NAMESPACE, "person", required=True)
        }



class Person(GenericObject):
    fields = {
        "name": StringField(),
        "cats": ReferenceFrom(NAMESPACE, "cat", "owner_id", cascade_delete=True),
        }



import inspect
import os
from slimit import ast
from slimit.parser import Parser

from fam.blud import GenericObject, ReferenceFrom
from fam.schema.validator import ModelValidator
from fam.buffer.buffer_views import FamWriteBufferViews

VIEW_FUNCTION_NAMES = ["map", "reduce"]


class ClassMapper(object):

    def __init__(self, classes, modules=None, designs=None):

        input_modules = modules if modules else []

        self.allow_public_write_types = []
        self.immutable_fields = {}
        self.namespaces = {}
        self.modules = {}
        self._add_classes(classes)
        self._add_modules(input_modules)
        self.sub_class_lookup = {}
        self._work_out_sub_classes()
        self.design_js_paths = designs if designs is not None else []
        self._buffer_views = None


    def extra_design_docs(self):

        docs = []
        for filepath in self.design_js_paths:
            design_doc = self._js_design_as_doc(filepath)
            docs.append(design_doc)

        return docs


    @property
    def buffer_views(self):
        if self._buffer_views is None:
            self._buffer_views = FamWriteBufferViews(self)
        return self._buffer_views


    def _add_immutable_field(self, type_name, field_name):

        # print "immutable", type_name, field_name
        field_names = self.immutable_fields.get(type_name)
        if field_names is None:
            field_names = []
            self.immutable_fields[type_name] = field_names
        field_names.append(field_name)


    def _add_modules(self, modules):
        for module in modules:
            classes = []
            for k, obj in module.__dict__.items():
                if inspect.isclass(obj):
                    if issubclass(obj, GenericObject):
                        if obj != GenericObject:
                            if not k.startswith("_"):
                                classes.append(obj)

            self._add_classes(classes)


    def __iter__(self):
        for name_space_name, name_space_classes in self.namespaces.items():
            for cls_name, cls in name_space_classes.items():
                yield cls


    def get_sub_class_names(self, namespace, class_name):
        return self.sub_class_lookup[(namespace, class_name)]


    def _work_out_sub_classes(self):
        # for each class add their subclasses type names to a lookup table keyed by namespace and classname
        # only works within a given namespace
        for namespace_name, namespace in self.namespaces.items():
            for class_name_super, cls_super in namespace.items():
                subclasses = []
                self.sub_class_lookup[(namespace_name, class_name_super)] = subclasses
                for class_name_sub, cls_sub in namespace.items():
                    if issubclass(cls_sub, cls_super):
                        subclasses.append(class_name_sub)


    def _add_classes(self, classes):

        for cls in classes:

            namespace_name = cls.namespace

            type_name = cls.type
            # gathers up information that gets added to the sync_gateway function
            if cls.sg_allow_public_write:
                self.allow_public_write_types.append(type_name)
            for field_name, field in cls.fields.items():
                if field.immutable:
                    self._add_immutable_field(type_name, field_name)
            namespace = self.namespaces.get(namespace_name)
            if namespace is None:
                namespace = {}
                self.namespaces[namespace_name] = namespace

            if not issubclass(cls, GenericObject):
                raise Exception("Classes you add to a ClassMapper must inherit from fam.blud.GenericObject this one does not: %s" % cls)
            namespace[type_name] = cls


    def get_class(self, type_name, namespace_name):
        namespace = self.namespaces.get(namespace_name)
        if namespace is None:
            return None
        return namespace.get(type_name)


    def _js_design_as_doc(self, filepath):

        dir, filename = os.path.split(filepath)
        name, ext = os.path.splitext(filename)

        with open(filepath) as f:
            js = f.read()

        parser = Parser()
        tree = parser.parse(js)

        views = {}

        for node in tree:
            if isinstance(node, ast.VarStatement):
                for child in node.children():
                    for grandchild in child.children():
                        if isinstance(grandchild, ast.Identifier):
                            view = {}
                            view_name = grandchild.value
                            views[view_name] = view
                        if isinstance(grandchild, ast.Object):
                            for named in grandchild.children():
                                function_name = None
                                function_body = None
                                for kv in named.children():
                                    if isinstance(kv, ast.Identifier) and kv.value in VIEW_FUNCTION_NAMES:
                                        function_name = kv.value
                                    if isinstance(kv, ast.FuncExpr):
                                        function_body = kv.to_ecma()
                                if function_name and function_body:
                                    view[function_name] = function_body


        return {"_id": "_design/%s" % name,
                "views": views}



    def get_design(self, namespace, namespace_name, foreign_key_str):

        views = {}
        for type_name, cls in namespace.items():
            for field_name, field in cls.cls_fields.items():
                if isinstance(field, ReferenceFrom):
                    view_key = "%s_%s" % (type_name, field_name)
                    # if view_key in ["person_dogs", "person_animals"]:
                    views[view_key] = {"map" : self._get_fk_map(field.refcls, field.refns, field.fkey, foreign_key_str)}

                if field.unique:
                    view_key = "%s_%s" % (type_name, field_name)
                    # if view_key in ["person_dogs", "person_animals"]:
                    views[view_key] = {"map": self._get_fk_map(type_name, namespace_name, field_name, foreign_key_str)}

        design = {
           "views": views
        }

        return design


    def get_all_subclass_names(self, namespace, class_name):

        if isinstance(class_name, list):
            class_names = class_name
        else:
            class_names = [class_name]

        all_sub_class_names = set()
        for name in class_names:
            sub_classes = set(self.get_sub_class_names(namespace, name))
            all_sub_class_names = all_sub_class_names.union(sub_classes)

        return all_sub_class_names



    def _get_fk_map(self, class_name, namespace, ref_to_field_name, foreign_key_str):

        all_sub_class_names = self.get_all_subclass_names(namespace, class_name)

        arrayStr = '["%s"]' % '", "'.join(all_sub_class_names)
        return foreign_key_str % (arrayStr, namespace, ref_to_field_name)






import unittest
import time
import sys
import subprocess
import requests
import json
from fam.utils import couchbase_utils

from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monarch, NAMESPACE
from fam.mapper import ClassMapper

from fam.database import CouchDBWrapper, SyncGatewayWrapper




class ReplicationTests(unittest.TestCase):

        def setUp(self):

            mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monarch])
            self.db_a = SyncGatewayWrapper(mapper,  "http://localhost:5985" , "sync_gateway")
            # self.db_b = SyncGatewayWrapper(mapper, "http://localhost:4985", "sync_gateway")


        # def test_put_some_dogs(self):
        #     dog_b1 = Dog(key="dog_b1", name="b1", channels=["shared_docs"])
        #     self.db_b.put(dog_b1)
        #     dog_b2 = Dog(key="dog_b2", name="b2", channels=["shared_docs"])
        #     self.db_b.put(dog_b2)
        #     dog_b3 = Dog(key="dog_b3", name="b3", channels=[])
        #     self.db_b.put(dog_b3)
        #
        #
        #     dog_a1 = Dog(key="dog_a1", name="a1", channels=["shared_docs"])
        #     self.db_a.put(dog_a1)
        #     dog_a2 = Dog(key="dog_a2", name="a2", channels=["shared_docs"])
        #     self.db_a.put(dog_a2)
        #     dog_a3 = Dog(key="dog_a3", name="a3", channels=[])
        #     self.db_a.put(dog_a3)
        #     dog_a4 = Dog(key="dog_a4", name="a4", channels=[])
        #     self.db_a.put(dog_a4)
        #     dog_a5 = Dog(key="dog_a5", name="a5", channels=[])
        #     self.db_a.put(dog_a5)
        #     dog_a6 = Dog(key="dog_a6", name="a6", channels=[])
        #     self.db_a.put(dog_a6)
        #
        #
        #     """
        #     a has a1-6 b1-2 seq 9
        #     b has a1-2 b1-3 seq 6 8556f2328ad7a3e626a5d4498a122ab7f33824e3 last 5
        #
        #
        #
        #     --- flush
        #
        #     a has a1-2 b1-2 seq 5
        #
        #
        #
        #
        #     """




        #
        #
        def test_put_some_more_dogs(self):

            dog_a7 = Dog(key="dog_a7", name="a7", channels=["shared_docs"])
            self.db_a.put(dog_a7)

        #
        #     """
        #
        #     do this offline
        #
        #     a a7 seq 2
        #
        #     turn online
        #
        #     a a1 a2 a7 b1 b2 seq 6
        #     b has a1-2 b1-3 seq 6 81ab002d844950713c06b88d404ef4571cefb04e last 6
        #
        #     """



        # #
        # def test_put_some_more_dogs(self):
        #
        #     dog_a8 = Dog(key="dog_a8", name="a8", channels=["shared_docs"])
        #     self.db_a.put(dog_a8)



import os
import json

from .requirement import CreateRequirement, UpdateRequirement, DeleteRequirement

THIS_DIR = os.path.dirname(__file__)


def _access_from_mapper(mapper):

    types = []
    for ns_name, ns in mapper.namespaces.items():
        for class_name, cls in ns.items():
            if cls.grants_access:
                types.append(class_name)
    return types


def _requirements_from_mapper(mapper):

    create_reqs = {}
    update_reqs = {}
    delete_reqs = {}

    requirements = {"create": create_reqs,
                    "update": update_reqs,
                    "delete": delete_reqs}

    for ns_name, ns in mapper.namespaces.items():
        for class_name, cls in ns.items():
            if cls.acl is not None:
                class_create_req = [r for r in cls.acl if isinstance(r, CreateRequirement)]
                if len(class_create_req) == 0:
                    create_reqs[class_name] = CreateRequirement().as_json()
                elif len(class_create_req) == 1:
                    create_reqs[class_name] = class_create_req[0].as_json()
                else:
                    raise Exception("too many create requirements in %s" % class_name)

                class_update_req = [r for r in cls.acl if isinstance(r, UpdateRequirement)]
                if len(class_update_req) == 0:
                    update_reqs[class_name] = [UpdateRequirement().as_json()]
                else:
                    update_reqs[class_name] = [req.as_json() for req in class_update_req]

                class_delete_req = [r for r in cls.acl if isinstance(r, DeleteRequirement)]
                if len(class_delete_req) == 0:
                    delete_reqs[class_name] = DeleteRequirement().as_json()
                elif len(class_create_req) == 1:
                    delete_reqs[class_name] = class_delete_req[0].as_json()
                else:
                    raise Exception("too many delete requirements in %s" % class_name)

    return requirements


def write_sync_function(config_template_path, sync_template_path, mapper, output_path, sync_function_symbol="SYNC_FUNCTION"):

    with open(config_template_path, "r") as f:
        config_src_str = f.read()

    config_src_str = config_src_str.replace("sync = ", "")

    with open(sync_template_path, "r") as f:
        permissions = f.read()

    requirements_str = json.dumps(_requirements_from_mapper(mapper))
    access_types_str = json.dumps(_access_from_mapper(mapper))

    permissions = permissions.replace("sync = ", "")
    permissions = permissions.replace('"REQUIREMENTS_LOOKUP"', requirements_str)
    permissions = permissions.replace('"ACCESS_TYPES"', access_types_str)

    config_str = config_src_str.replace(sync_function_symbol, permissions)

    with open(output_path, "w") as f:
        f.write(config_str)









"""
This is just work in progress

Doesnt do anything right now
"""



import datetime


class FamMigration(object):

    timestamp = "2016-05-30T09:30:10Z"

    def up(self, db, doc):
        pass

    def down(self, db, doc):
        pass


def migrate(db, obj, migrations, ts=None):

    dst = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ") if ts is None else ts



import js2py
import json

class FamWriteBufferViews(object):

    FOREIGN_KEY_MAP_STRING = '''function(doc) {
        var resources = %s;
        if (resources.indexOf(doc.type) != -1 && doc.namespace == \"%s\"){
            emit(doc.%s, doc);
        }
    }'''

    def __init__(self, mapper):

        self.mapper = mapper
        self.views = {}
        self.indexes = {}
        self.reverse_indexes = {}
        self.js_context = None
        self._update_designs()
        self.view_name = None
        self.doc = None


    def clear_indexes(self):
        self.indexes = {}
        self.reverse_indexes = {}


    def query_view(self, view_name, key, **kwargs):

        k = tuple(key) if type(key) == list else key

        name = self._clean_name(view_name)
        view_index = self.indexes.get(name)
        # print "****************** - ", self.indexes.get("join_views_reel_item_memberships")
        if view_index is None:
            return []
        values = view_index.get(k)
        if values is None:
            return []
        return [item[1] for item in values.items()]


    def index_obj(self, obj):

        self.obj = obj
        doc = obj.as_dict()
        for view_name, view in self.views.items():
            self.view_name = view_name
            view(doc)


    def _clean_name(self, name):
        return name.replace("/", "_").replace(".", "_").replace("-", "_").replace(":", "_")


    def remove_from_indexes(self, obj_id):

        for view_name in self.indexes.keys():
            index = self.indexes.get(view_name)
            reverse_index = self.reverse_indexes.get(view_name)

            existing_key = reverse_index.get(obj_id)

            ## remove the previous entry
            if existing_key is not None:
                old_indexed_values = index[existing_key]
                del old_indexed_values[obj_id]
                del reverse_index[obj_id]



    def _add_to_index(self, k):

        # print "********* adding: ", k

        kstr = repr(k).replace("'", "\"").replace("u\"", "\"")
        key = None if kstr == '"undefined"' else json.loads(kstr)

        if type(key) == list:
            key = tuple(key)
            # print "************ adding key: ", key

        index = self.indexes.get(self.view_name)
        reverse_index = self.reverse_indexes.get(self.view_name)
        if index is None:
            index = {}
            self.indexes[self.view_name] = index

        if reverse_index is None:
            reverse_index = {}
            self.reverse_indexes[self.view_name] = reverse_index

        obj_id = self.obj.key
        existing_key = reverse_index.get(obj_id)

        if existing_key == key:
            return

        ## remove the previous entry
        if existing_key is not None:
            old_indexed_values = index[existing_key]
            del old_indexed_values[obj_id]
            del reverse_index[obj_id]

        ## add the new one
        if key is not None:
            reverse_index[obj_id] = key
            new_indexed_values = index.get(key)
            if new_indexed_values is None:
                new_indexed_values = {}
                index[key] = new_indexed_values
            new_indexed_values[obj_id] = self.obj


    def _raw_design_doc(self):
        design_doc = {
            "views": {
                "all": {
                    "map": "function(doc) {emit(doc.type, doc);}"
                }
            }
        }
        return design_doc


    def _update_designs(self):

        def add(k):
            self._add_to_index(k)

        self.js_context = js2py.EvalJs({"add": add})

        code = """
        function emit(k, v){
            add(k);
        }
        """

        self.js_context.execute(code)

        ## simple type index
        doc = self._raw_design_doc()
        key = "_design/raw"
        doc["_id"] = key

        self._add_design(self.js_context, key, doc)

        # ## relational indexes
        for namespace_name, namespace in self.mapper.namespaces.items():
            view_namespace = self._clean_name(namespace_name)
            key = "_design/%s" % view_namespace
            doc = self.mapper.get_design(namespace, namespace_name, self.FOREIGN_KEY_MAP_STRING)
            doc["_id"] = key
            self._add_design(self.js_context, key, doc)

        # ## extra indexes
        for doc in self.mapper.extra_design_docs():
            key = doc["_id"]
            self._add_design(self.js_context, key, doc)


    def _add_design(self, js_context, key, doc):

        if(not key.startswith("_design/")):
            raise Exception("DataBaseCacheViews design doc key should start with _design")

        design_name = key[len("_design/"):]

        for view_name, view in doc["views"].items():
            name = "%s_%s" % (design_name, view_name)
            name = self._clean_name(name)
            code = "var %s = %s" % (name, view["map"])
            js_context.execute(code)
            self.views[name] = getattr(js_context, name)



import os

import atexit
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install


def delete_old_slimit_files():
    """
    slimit generates two files with code generation lextab.py and yacctab.py
    slimit install ships with out of date versions of these files that have been created by ply 3.4
    To use ply 3.8 with slimit 0.8.1 you have to delete thses two files, they will be lazily recreated by slimit using the up to date ply
    """
    import site
    from importlib import reload
    reload(site)
    import sys
    def find_module_path():
        for p in sys.path:
            if os.path.isdir(p) and "slimit" in os.listdir(p):
                return os.path.join(p, "slimit")

    install_path = find_module_path()

    file_names = ["lextab.py", "yacctab.py"]
    for file_name in file_names:
        file_path = os.path.join(install_path, file_name)
        if os.path.exists(file_path):
            os.remove(file_path)


class PostDevelopCommand(develop, object):
    """Post-installation for development mode."""

    def __init__(self, *args, **kwargs):
        super(PostDevelopCommand, self).__init__(*args, **kwargs)
        def _post_install():
            delete_old_slimit_files()
        atexit.register(_post_install)


class PostInstallCommand(install, object):
    """Post-installation for installation mode."""

    def __init__(self, *args, **kwargs):
        super(PostInstallCommand, self).__init__(*args, **kwargs)
        def _post_install():
            delete_old_slimit_files()
        atexit.register(_post_install)


setup(name='fam',
    version='2.0.10',
    description="Simple Python ORM for CouchDB, and Sync Gateway",
    url="https://github.com/paulharter/fam",
    classifiers=[
      'Development Status :: 4 - Beta',
      'Natural Language :: English',
      'Operating System :: OS Independent',
      'Programming Language :: Python :: 3.6',
      'License :: OSI Approved :: MIT License'
    ],
    author='Paul Harter',
    author_email='paul@glowinthedark.co.uk',
    license="LICENSE",
    install_requires=['js2py', 'requests', 'simplejson', 'jsonschema', 'mock', 'pytz', 'slimit', 'ply==3.8',
                    'firebase_admin', 'six'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    cmdclass={
      'develop': PostDevelopCommand,
      'install': PostInstallCommand,
    }
)




CREATE = "create"
UPDATE = "update"
DELETE = "delete"



class BaseRequirement(object):

    action = None

    def __init__(self, user=None, role=(), access=True, owner=False):

        self.user = user
        self.role = role
        self.access = access
        self.owner = owner

    def as_json(self):

        j = {}
        if self.user is not None:
            j["user"] = self.user
        if self.role is not None:
            j["role"] = self.role
        if self.access is False:
            j["withoutAccess"] = True
        if self.owner is True:
            j["owner"] = True

        return j


class CreateRequirement(BaseRequirement):
    pass


class UpdateRequirement(BaseRequirement):

    def __init__(self, user=None, role=(), access=True, owner=False, fields=None):

        super(UpdateRequirement, self).__init__(user=user, role=role, access=access, owner=owner)
        self.fields = fields

    def as_json(self):

        j = super(UpdateRequirement, self).as_json()
        if self.fields is not None:
            j["fields"] = self.fields
        return j


class DeleteRequirement(BaseRequirement):
    pass






import os

import unittest
from fam.database import CouchbaseWrapper
from fam.mapper import ClassMapper

from fam.tests.models.test01 import Dog, Cat

COUCHBASE_HOST = "127.0.0.1"
COUCHBASE_BUCKET = "test"


THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class CacheTests(unittest.TestCase):

    def setUp(self):
        mapper = ClassMapper([Dog, Cat])
        self.db = CouchbaseWrapper(mapper, COUCHBASE_HOST, COUCHBASE_BUCKET, read_only=False)

    def tearDown(self):
        pass

    def test_save(self):
        dog = Dog(name="fly")
        dog.save(self.db)

    def test_get(self):
        dog = Dog(name="fly")
        dog.save(self.db)
        dog_key = dog.key
        got_dog = self.db.get(dog_key)
        self.assertEqual(dog.name, got_dog.name)

    def test_n1ql(self):
        dog = Dog(name="fly")
        dog.save(self.db)
        dogs = self.db.n1ql('SELECT META(test).cas, META(test).id, * FROM test WHERE type="dog" and name="fly"')



import sys
import datetime
import pytz
from fam.extra_types.lat_long import LatLong
from fractions import Fraction
from decimal import Decimal
from google.cloud.firestore_v1beta1 import GeoPoint
import base64

from .base_adapter import BaseDataAdapter


if sys.version_info[0] < 3:
    PYTHON_VERSION = 2
else:
    PYTHON_VERSION = 3


class CouchDBDataAdapter(BaseDataAdapter):
    pass





import unittest
import json
import time
import os


from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE
from fam.tests.models.test04 import House

from fam.database import FirestoreWrapper
from fam.database import CouchDBWrapper
from fam.extra_types.lat_long import LatLong

from fam.mapper import ClassMapper

from fam.firestore_sync.syncer import FirestoreSyncer

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))))


class TestDB(unittest.TestCase):


    def setUp(self):
        creds_path = os.path.join(ROOT_DIR, "secrets", "earth-rover-test-d241bce5266d.json")
        mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey])
        self.firestore = FirestoreWrapper(mapper, creds_path, namespace=NAMESPACE)
        self.couchdb = CouchDBWrapper(mapper, "http://localhost:5984", db_name="test", reset=True)
        self.couchdb.update_designs()
        self.clear_db()


    def clear_db(self):
        self.firestore.delete_all("dog")
        self.firestore.delete_all("cat")
        self.firestore.delete_all("person")
        self.firestore.delete_all("jackrussell")
        self.firestore.delete_all("monkey")


    def test_app(self):
        self.assertNotEqual(self.firestore, None)


    def test_query_generator(self):

        paul = Person.create(self.firestore, name="paul")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)
        dog2 = Dog.create(self.firestore, name="tiny", owner=paul)
        dog3 = Dog.create(self.firestore, name="fly", owner=paul)

        dogs_ref = self.firestore.db.collection("dog")
        q = dogs_ref.where(u"owner_id", u"==", paul.key)

        dogs = self.firestore.query_items(q, batch_size=1)
        dogs_list = list(dogs)
        self.assertTrue(isinstance(dogs_list[0], Dog))

        self.assertEquals(len(dogs_list), 3)


    def test_sync_down(self):

        paul = Person.create(self.firestore, name="paul")
        sol = Person.create(self.firestore, name="sol")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)
        dog2 = Dog.create(self.firestore, name="tiny", owner=paul)
        dog3 = Dog.create(self.firestore, name="fly", owner=paul)

        dogs_ref = self.firestore.db.collection("dog")

        syncer = FirestoreSyncer(self.couchdb, self.firestore)
        syncer.add_query(dogs_ref.where("owner_id", "==", paul.key))

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 0)

        syncer.sync_down()

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 3)


    def test_sync_down_since(self):

        paul = Person.create(self.firestore, name="paul")
        sol = Person.create(self.firestore, name="sol")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)
        dog2 = Dog.create(self.firestore, name="tiny", owner=paul)
        dog3 = Dog.create(self.firestore, name="fly", owner=paul)

        dogs_ref = self.firestore.db.collection("dog")

        syncer = FirestoreSyncer(self.couchdb, self.firestore)
        syncer.add_query(dogs_ref.where("owner_id", "==", paul.key))

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 0)

        changed = syncer.sync_down()
        self.assertEqual(len(changed), 3)

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 3)

        dog3.update({"name":"jelly"})

        changed = syncer.sync_down()
        print(changed)
        self.assertEqual(len(changed), 1)


        updated = self.couchdb.get(dog3.key)

        self.assertEqual(updated.name, "jelly")


    def test_sync_down_since_in_db(self):

        paul = Person.create(self.firestore, name="paul")
        sol = Person.create(self.firestore, name="sol")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)
        dog2 = Dog.create(self.firestore, name="tiny", owner=paul)
        dog3 = Dog.create(self.firestore, name="fly", owner=paul)

        dogs_ref = self.firestore.db.collection("dog")

        syncer = FirestoreSyncer(self.couchdb, self.firestore, since_in_db=True)
        syncer.add_query(dogs_ref.where("owner_id", "==", paul.key))

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 0)

        changed = syncer.sync_down()
        self.assertEqual(len(changed), 3)

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 3)

        dog3.update({"name":"jelly"})

        changed = syncer.sync_down()
        print(changed)
        self.assertEqual(len(changed), 1)


        updated = self.couchdb.get(dog3.key)

        self.assertEqual(updated.name, "jelly")



    def test_sync_down_single(self):

        paul = Person.create(self.firestore, name="paul")
        sol = Person.create(self.firestore, name="sol")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)

        syncer = FirestoreSyncer(self.couchdb, self.firestore)

        dog1_ref = self.firestore.db.collection("dog").document(dog1.key)

        syncer.add_doc_ref(dog1_ref)

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 0)

        syncer.sync_down()
        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 1)



    def test_sync_up(self):

        paul = Person.create(self.firestore, name="paul")
        sol = Person.create(self.firestore, name="sol")
        dog1 = Dog.create(self.firestore, name="woofer", owner=paul)
        dog2 = Dog.create(self.firestore, name="tiny", owner=paul)
        dog3 = Dog.create(self.firestore, name="fly", owner=paul)

        dogs_ref = self.firestore.db.collection("dog")

        syncer = FirestoreSyncer(self.couchdb, self.firestore)
        syncer.add_query(dogs_ref.where("owner_id", "==", paul.key))

        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 0)

        syncer.sync_down()
        dogs = Dog.all(self.couchdb)
        dogs_list = list(dogs)
        self.assertEqual(len(dogs_list), 3)

        dog4 = Dog.create(self.couchdb, name="di", owner_id=paul.key)
        dog5 = Dog.create(self.couchdb, name="stevie", owner_id=paul.key)

        syncer.sync_up()
        dogs = list(paul.dogs)
        self.assertEqual(len(dogs), 5)




